from zipfile import ZipFile
from io import BytesIO
import logging

from construct import *
from simpleelf.elf_structs import ElfStructs
from simpleelf import elf_consts

from packers.utils import guess_machine_type

COMPRESSED_SECTION_MAGIC = b'\xfe\xed\xfa\xce'

elf_compressed_section = Struct(
    '_start' / Tell,
    'magic' / Const(COMPRESSED_SECTION_MAGIC),
    'unknown0' / Int32ub,
    'size' / Hex(Int32ub),
    'unknown2' / Int32ub,
    'unknown3' / Int32ub,
    'data' / Bytes(this.size),
)

elf_structs = ElfStructs('>')

def extract_elf(fw):
    header_magic = elf_consts.ELFMAG
    is_elf = header_magic == fw.read(len(header_magic))
    fw.seek(0)

    if not is_elf:
        return None

    parsed_elf = elf_structs.Elf32.parse_stream(fw)
    for s in parsed_elf.sections:
        if s.data is None:
            continue

        if not s.data.startswith(COMPRESSED_SECTION_MAGIC):
            continue

        compressed_elf = elf_compressed_section.parse(s.data)
        with ZipFile(BytesIO(compressed_elf.data)) as z:
            entries = z.namelist()

            if len(entries) != 1:
                raise Exception("invalid zip archive entry")

            return z.open(entries[0]).read()

    return None


def unpack(fw, out):
    logging.debug('attempting ELF packer')

    raw_elf = extract_elf(fw)

    if raw_elf is None:
        logging.debug('failed to extract ELF')
        return False

    logging.info('ELF image format. extracting...')

    parsed_elf = elf_structs.Elf32.parse(raw_elf)
    parsed_elf.header.e_machine = guess_machine_type(raw_elf)
    out.write(elf_structs.Elf32.build(parsed_elf))

    logging.info('extracted successfully')
    return True



import bz2
import logging

from construct import *

from simpleelf.elf_structs import ElfStructs
from simpleelf.elf_builder import ElfBuilder
from simpleelf import elf_consts

from packers.utils import CRASH_MAGIC, get_elf_from_code

elf_structs = ElfStructs('>')

def unpack(fw, out):
    logging.debug('attempting RAW packer')

    main_image = fw.read()

    if CRASH_MAGIC not in main_image:
        logging.debug('didnt find CRASH_MAGIC')
        return False

    e = get_elf_from_code(main_image)
    out.write(e.build())

    logging.info('extracted successfully')
    return True



import bz2
import logging

from construct import *

MAGIC = b'RV'

rv_struct = Struct(
    'magic' / Const(MAGIC),
    Seek(0x2c),
    'elf_size' / Int32ub,

    'unknown0' / Int32ub,
    'unknown1' / Int32ub,
    'unknown2' / Int32ub,
    'unknown3' / Int32ub,

    'elf' / Bytes(this.elf_size),
)

def unpack(fw, out):
    logging.debug('attempting RV packer')
    main_image = fw.read()

    if not main_image.startswith(MAGIC):
        logging.debug('not RV image')
        return False

    rv = rv_struct.parse(main_image)
    out.write(rv.elf)

    logging.info('extracted successfully')
    return True



import logging

from construct import *

from simpleelf.elf_structs import ElfStructs
from simpleelf.elf_builder import ElfBuilder
from simpleelf import elf_consts

CRASH_MAGIC = b'\xDE\xAD\x12\x34'

crashinfo_external = Struct(
    'magic' / Const(CRASH_MAGIC),

    'version' / Int32ub,
    'reason' / Int32ub,
    'cpu_vector' / Int32ub,
    'registers' / Int32ub,
    'rambase' / Int32ub,

    'textbase' / Hex(Int32ub),
    'database' / Hex(Int32ub),
    'bssbase' / Hex(Int32ub),
    'heapbase' / Hex(Int32ub),

)

def guess_machine_type(code):
    # assuming cisco only uses mips and ppc cpus,
    # check for: isync; sync; should be enough to 
    # determine the cpu type
    if b'\x4C\x00\x01\x2C\x7C\x00\x04\xAC' in code:
        return elf_consts.EM_PPC
    else:
        return elf_consts.EM_MIPS

def get_elf_from_code(main_image, entry=None):
    e = ElfBuilder()
    e.set_endianity('>')

    crashinfo_external_offset = main_image.find(CRASH_MAGIC)
    if -1 == crashinfo_external_offset:
        logging.warning("failed to locate regions magic. using single .text section")
        if entry is None:
            entry = 0
        e.add_segment(entry, main_image, 
            elf_consts.PF_R | elf_consts.PF_W | elf_consts.PF_X)
        e.add_code_section('.text', entry, len(main_image))
    else:
        regions = crashinfo_external.parse(main_image[crashinfo_external_offset:
            crashinfo_external_offset+crashinfo_external.sizeof()])

        if entry is None:
            entry = regions.textbase & 0xfffff000

        e.add_segment(entry, main_image, 
            elf_consts.PF_R | elf_consts.PF_W | elf_consts.PF_X)
        
        e.add_code_section('.text', regions.textbase, len(main_image) - (regions.textbase - entry))
        e.add_empty_data_section('.data', regions.database, regions.bssbase - regions.database)
        e.add_empty_data_section('.bss', regions.bssbase, regions.heapbase - regions.bssbase)

    e.set_entry(entry)
    e.set_machine(guess_machine_type(main_image))

    return e
